#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

import logging
from typing import Optional

logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.DEBUG)


class SlipMixin:
    """
    The ``SlipMixin`` helps to encode and decode frames transferred
    in the Serial Line Internet Protocol.

    It cannot be used standalone and must be mixed into a class that
    implements the :class:`htf.communication.SlipInterface`.
    """

    END = b"\xc0"  # 0o300
    ESC = b"\xdb"  # 0o333
    ESC_END = b"\xdc"  # 0o334
    ESC_ESC = b"\xdd"  # 0o335

    def slip_send_frame(self, frame: bytes) -> None:
        """
        Send a frame encoded in the Serial Line Internet Protocol.

        Args:
            frame: the frame to be sent.
        """
        assert isinstance(frame, bytes)

        # send initial END to flush the receiver
        self.send_byte(self.END)  # type: ignore

        for byte in [frame[i : i + 1] for i in range(len(frame))]:
            if byte == self.END:
                # escape an end byte
                self.send_byte(self.ESC)  # type: ignore
                self.send_byte(self.ESC_END)  # type: ignore

            elif byte == self.ESC:
                # escape an escape byte
                self.send_byte(self.ESC)  # type: ignore
                self.send_byte(self.ESC_ESC)  # type: ignore

            else:
                # else simply send the current byte
                self.send_byte(byte)  # type: ignore

        self.send_byte(self.END)  # type: ignore

    def slip_receive_frame(self, timeouts: Optional[int] = None, max_length: int = 1024) -> Optional[bytes]:
        """
        Receive a frame encoded in the Serial Line Internet Protocol
        and return the decoded frame.

        Args:
            timeouts=None: the number of timeouts until None is returned.
                The timeout depends on the mixing class and underlying
                communication system.
            max_length=1024: maximum number of bytes that will be
                received, not including ESC. Prevents the method from
                perpetually blocking.

        Returns:
            bytes or None: the decoded frame if a frame was received or
                ``None`` if no frame was received.
        """

        received = b""
        byte = b""
        number_of_timeouts = 0
        bytes_received = 0

        while timeouts is None or number_of_timeouts < timeouts:
            byte = self.receive_byte()  # type: ignore

            if byte in (None, b""):
                number_of_timeouts += 1
                continue

            if byte == self.END or bytes_received > max_length:
                if received:
                    # return a frame if at least one byte was received
                    return received

            elif byte == self.ESC:
                byte = self.receive_byte()  # type: ignore
                if byte in (None, b""):
                    number_of_timeouts += 1
                    continue

                if byte == self.ESC_END:
                    # append END
                    received += self.END

                elif byte == self.ESC_ESC:
                    # append ESC
                    received += self.ESC
                else:
                    # protocol violation
                    # stuff violating byte into the packet anyway
                    received += byte
                    print("protocol vialation")
            else:
                received += byte
            bytes_received += 1
        return None


class SlipInterface(SlipMixin):  # pragma: no cover
    def send_byte(self, byte: bytes) -> None:
        """
        Send a single byte.

        Args:
            byte: a single byte to be sent.
        """
        assert False

    def receive_byte(self) -> bytes:
        """
        Receive a single byte. If no byte was received, ``None`` is returned to
        tell about a timeout.

        Returns:
            bytes or None: a single byte that was received or ``None`` if a
                timeout occurred.
        """
        assert False
